import * as React from "react";
import { Rect } from "../Rect";
import { CLASSES } from "../Types";

/** @internal */
export interface IFloatingWindowProps {
  title: string;
  id: string;
  url: string;
  rect: Rect | null;
  onCloseWindow: (id: string) => void;
  onSetWindow: (id: string, window: Window) => void;
}

interface IStyleSheet {
  href: string | null;
  type: string;
  rules: string[] | null;
}

/** @internal */
export const FloatingWindow = (
  props: React.PropsWithChildren<IFloatingWindowProps>
) => {
  const { title, id, url, rect, onCloseWindow, onSetWindow } = props;
  const popoutWindow = React.useRef<Window | null>(null);
  const timerId = React.useRef<any>(null);
  const contentRef = React.useRef<HTMLDivElement | null>(null);

  React.useLayoutEffect(() => {
    if (timerId.current) {
      clearTimeout(timerId.current);
    }
    let isMounted = true;
    const r = rect || new Rect(0, 0, 100, 100);
    const styles = Array.from(window.document.styleSheets).reduce(
      (result, styleSheet) => {
        let rules: CSSRuleList | undefined = undefined;
        try {
          rules = styleSheet.cssRules;
        } catch (e) {
          // styleSheet.cssRules can throw a security exception
        }

        try {
          return [
            ...result,
            {
              href: styleSheet.href,
              type: styleSheet.type,
              rules: rules ? Array.from(rules).map((rule) => rule.cssText) : null,
            },
          ];
        } catch (e) {
          return result;
        }
      },
      [] as IStyleSheet[]
    );
    popoutWindow.current = window.open(
      url,
      id,
      `left=${r.x},top=${r.y},width=${r.width},height=${r.height}`
    );
    if (popoutWindow.current !== null) {
      onSetWindow(id, popoutWindow.current);

      window.addEventListener("beforeunload", () => {
        if (popoutWindow.current) {
          popoutWindow.current.close();
          popoutWindow.current = null;
        }
      });

      popoutWindow.current.addEventListener("load", () => {
        if (isMounted) {
          const popoutDocument = popoutWindow.current!.document;
          popoutDocument.title = title;
          const popoutContent = popoutDocument.createElement("div");
          popoutContent.className = CLASSES.FLEXLAYOUT__FLOATING_WINDOW_CONTENT;
          popoutDocument.body.appendChild(popoutContent);
          contentRef.current = popoutContent;

          copyStyles(popoutDocument, styles).then(() => {
            // setContent(popoutContent); // Ne pas utiliser setContent avec createPortal désactivé
          });

          popoutWindow.current!.addEventListener("beforeunload", () => {
            onCloseWindow(id);
          });
        }
      });
    } else {
      console.warn(`Unable to open window ${url}`);
      onCloseWindow(id);
    }

    return () => {
      isMounted = false;
      timerId.current = setTimeout(() => {
        if (popoutWindow.current) {
          popoutWindow.current.close();
          popoutWindow.current = null;
        }
      }, 0);
    };
  }, []);

  // Utiliser contentRef.current comme nécessaire dans le reste du code...

  return null;
};

/** @internal */
function copyStyles(doc: Document, styleSheets: IStyleSheet[]): Promise<boolean[]> {
  const head = doc.head;
  const promises: Promise<boolean>[] = [];
  for (const styleSheet of styleSheets) {
    if (styleSheet.href) {
      const styleElement = doc.createElement("link");
      styleElement.type = styleSheet.type;
      styleElement.rel = "stylesheet";
      styleElement.href = styleSheet.href!;
      head.appendChild(styleElement);
      promises.push(
        new Promise((resolve) => {
          styleElement.onload = () => resolve(true);
        })
      );
    } else {
      if (styleSheet.rules) {
        const style = doc.createElement("style");
        for (const rule of styleSheet.rules) {
          style.appendChild(doc.createTextNode(rule));
        }
        head.appendChild(style);
      }
    }
  }
  return Promise.all(promises);
}
